import numpy as np
import matplotlib.pyplot as plt
from pysimenv.core.utils import Logger


class StateVariable:
    def __init__(self, state):
        self.state = np.array(state)
        self.deriv = np.zeros_like(state)
        self.rk4_buffer = []
        self.correction_fun = None

    def apply_state(self, state):
        self.state = np.array(state)

    @property
    def size(self):
        return self.state.size

    def forward(self, deriv):
        self.deriv = deriv

    def rk4_update_1(self, dt):
        x_0 = np.copy(self.state)
        k_1 = np.copy(self.deriv)

        self.rk4_buffer.clear()
        self.rk4_buffer.append(x_0)
        self.rk4_buffer.append(k_1)
        self.state = x_0 + dt / 2 * k_1  # x = x0 + dt/2*k1
        if self.correction_fun is not None:
            self.state = self.correction_fun(self.state)

    def rk4_update_2(self, dt):
        x_0 = self.rk4_buffer[0]
        k_2 = np.copy(self.deriv)

        self.rk4_buffer.append(k_2)
        self.state = x_0 + dt / 2 * k_2  # x = x0 + dt/2*k2
        if self.correction_fun is not None:
            self.state = self.correction_fun(self.state)

    def rk4_update_3(self, dt):
        x_0 = self.rk4_buffer[0]
        k_3 = np.copy(self.deriv)

        self.rk4_buffer.append(k_3)
        self.state = x_0 + dt / 2 * k_3  # x = x0 + dt/2*k3
        if self.correction_fun is not None:
            self.state = self.correction_fun(self.state)

    def rk4_update_4(self, dt):
        x_0, k_1, k_2, k_3 = self.rk4_buffer[:]
        k_4 = np.copy(self.deriv)

        self.state = x_0 + dt * (k_1 + 2 * k_2 + 2 * k_3 + k_4) / 6
        if self.correction_fun is not None:
            self.state = self.correction_fun(self.state)

    def attach_correction_fun(self, correction_fun):
        self.correction_fun = correction_fun

    @staticmethod
    def test():
        state_var = StateVariable(1)
        def deriv_fun(x): return -x
        dt = 0.01

        time_list = []
        sim_state_list = []
        t = 0
        for i in range(1000):
            time_list.append(t)
            sim_state_list.append(state_var.state)

            state_var.forward(deriv_fun(state_var.state))
            state_var.rk4_update_1(dt)
            state_var.forward(deriv_fun(state_var.state))
            state_var.rk4_update_2(dt)
            state_var.forward(deriv_fun(state_var.state))
            state_var.rk4_update_3(dt)
            state_var.forward(deriv_fun(state_var.state))
            state_var.rk4_update_4(dt)
            t = t + dt

        time_list = np.array(time_list)
        sim_state_list = np.array(sim_state_list)
        true_state_list = np.exp(-time_list)

        plt.figure()
        plt.subplot(2, 1, 1)
        plt.plot(time_list, sim_state_list, color='k')
        plt.xlabel("Time [s]")
        plt.ylabel("State")
        plt.grid()

        plt.subplot(2, 1, 2)
        plt.plot(time_list, sim_state_list - true_state_list, color='k')
        plt.xlabel("Time [s]")
        plt.ylabel("Integration error")
        plt.grid()

        plt.subplots_adjust(bottom=0.1, left=0.1, top=0.9, hspace=0.5)
        plt.show()


class SimObject(object):
    FLAG_OPERATING = 0

    def __init__(self):
        self.flag = SimObject.FLAG_OPERATING
        self.sim_clock = None
        self.log_timer = None
        self.logger = Logger()

    def attach_sim_clock(self, sim_clock):
        self.sim_clock = sim_clock

    def attach_log_timer(self, log_timer):
        self.log_timer = log_timer

    def detach_sim_clock(self):
        self.sim_clock = None

    def detach_log_timer(self, log_timer):
        self.log_timer = None
    
    def reset(self):
        self.flag = SimObject.FLAG_OPERATING
        self.logger.clear()

    @property
    def time(self):
        assert self.sim_clock is not None, "[BaseFunction] Attach a sim_clock first!"
        return self.sim_clock.time

    # to be implemented
    def forward(self):
        return

    # to be implemented
    def check_stop_condition(self):
        to_stop = False
        flag = self.flag
        return to_stop, flag


if __name__ == "__main__":
    StateVariable.test()