import numpy as np
import matplotlib.pyplot as plt
from pytictoc import TicToc
from pysimenv.core.utils import SimClock
from pysimenv.core.utils import Timer
from pysimenv.core.systems import TimeInvarDynSystem


class Simulator(object):
    def __init__(self, model):
        self.sim_clock = SimClock()
        self.log_timer = Timer(np.inf)
        self.log_timer.attach_sim_clock(self.sim_clock)

        self.model = model
        self.model.attach_sim_clock(self.sim_clock)
        self.model.attach_log_timer(self.log_timer)

    def reset(self):
        self.sim_clock.reset()
        self.log_timer.turn_off()
        self.model.reset()

    def begin_logging(self, log_interval):
        self.log_timer.turn_on(log_interval)

    def finish_logging(self):
        self.log_timer.turn_off()

    def step(self, dt, *args, **kwargs):
        self.model.step(dt, *args, **kwargs)

    def propagate(self, dt, time, save_history=True, *args, **kwargs):
        if save_history:
            self.begin_logging(dt)

        print("[Simulator] Simulating...")
        tic_toc = TicToc()
        tic_toc.tic()
        self.model.propagate(dt, time, *args, **kwargs)
        elapsed_time = tic_toc.tocvalue()

        print("[Simulator] Elapsed time: {:.4f} [s]".format(elapsed_time))
        self.finish_logging()

    @staticmethod
    def test():
        def deriv_fun(x, u):
            A = np.array([[0, 1], [-1, -1]])
            B = np.array([0, 1])
            return A.dot(x) + B.dot(u)

        model = TimeInvarDynSystem([0, 1], deriv_fun)
        simulator = Simulator(model)
        simulator.propagate(dt=0.01, time=10, save_history=True, u=1)
        model.plot()

        simulator.reset()
        simulator.begin_logging(0.01)
        for i in range(1000):
            simulator.step(dt=0.01, u=1)
        simulator.finish_logging()
        model.plot()

        plt.show()


if __name__ == "__main__":
    Simulator.test()
