import numpy as np
import matplotlib.pyplot as plt
from pytictoc import TicToc
from pysimenv.core.utils import SimClock, Timer
from pysimenv.core.base import SimObject, StateVariable
from pysimenv.core.functions import BaseFunction


class BaseSystem(SimObject):
    def __init__(self, *initial_states):
        super().__init__()
        self.name = "base_system"
        self.state_var_list = []
        self.state_index = []

        last_index = 0
        if len(initial_states) > 0:
            for initial_state in initial_states:
                var = StateVariable(initial_state)
                index = list(range(last_index, last_index + var.size))
                self.state_var_list.append(var)
                self.state_index.append(index)
                last_index += var.size

        self.state_var_num = len(self.state_var_list)
        self.state_num = last_index

    def set_state(self, *states):
        if len(states) > 0:
            for i, state in enumerate(states):
                self.state_var_list[i].apply_state(state)

    # to be implemented
    def forward(self, *args, **kwargs):
        return

    def step(self, dt, *args, **kwargs):
        t_0 = self.sim_clock.time

        self.log_timer.forward()
        self.forward(*args, **kwargs)
        for var in self.state_var_list:
            var.rk4_update_1(dt)

        self.sim_clock.apply_time(t_0 + dt/2)
        self.log_timer.forward()
        self.forward(*args, **kwargs)
        for var in self.state_var_list:
            var.rk4_update_2(dt)

        self.forward(*args, **kwargs)
        for var in self.state_var_list:
            var.rk4_update_3(dt)

        self.sim_clock.apply_time(t_0 + dt - 10*self.sim_clock.time_res)
        self.log_timer.forward()
        self.forward(*args, **kwargs)
        for var in self.state_var_list:
            var.rk4_update_4(dt)

        self.sim_clock.apply_time(t_0 + dt)

    def propagate(self, dt, time, *args, **kwargs):
        assert self.sim_clock is not None, "Attach a sim_clock first!"
        assert self.log_timer is not None, "Attach a log_timer first!"

        iter_num = min(round(time/dt), np.iinfo(np.int32).max)
        for i in range(iter_num):
            to_stop, _ = self.check_stop_condition()
            if to_stop:
                break
            self.step(dt, *args, **kwargs)

        self.log_timer.forward()
        self.forward(*args, **kwargs)

    def history(self, *keys):
        return self.logger.get(*keys)

    # to be implemented
    @property
    def output(self):
        return

    def plot(self, var_keys=None, var_ind_dict=None, var_names_dict=None):
        if var_keys is None:
            var_keys = list(self.logger.keys())
            var_keys.remove('time')
        if var_ind_dict is None:
            var_ind_dict = dict()
        if var_names_dict is None:
            var_names_dict = dict()

        figs = []
        time_list = self.history('time')

        for var_key in var_keys:
            fig = plt.figure()
            figs.append(fig)

            var_list = self.history(var_key)
            if var_list.ndim == 1:
                var_list = np.reshape(var_list, var_list.shape + (1, ))

            if var_key in var_ind_dict:
                ind = var_ind_dict[var_key]
            else:
                ind = list(range(var_list.shape[1]))

            if var_key in var_names_dict:
                names = var_names_dict[var_key]
            else:
                names = [var_key + "[" + str(k) + "]" for k in ind]

            subplot_num = len(ind)
            for i in range(subplot_num):
                plt.subplot(subplot_num, 1, i + 1)
                plt.plot(time_list, var_list[:, ind[i]], label="Actual")
                plt.xlabel("Time (s)")
                plt.ylabel(names[i])
                plt.grid()
                plt.legend()
            plt.suptitle(var_key)

        plt.draw()
        plt.pause(0.01)
        return figs

    # to be implemented
    def report(self):
        print("== Report for {} ==".format(self.name))
        return

    def save(self, h5file=None, data_group='/'):
        data_group = data_group + self.name + '/'
        self.logger.save(h5file, data_group)

    def load(self, h5file=None, data_group='/'):
        data_group = data_group + self.name + '/'
        self.logger.load(h5file, data_group)

    @property
    def state(self):
        return self._get_state()

    def _get_state(self):
        states = []
        for var in self.state_var_list:
            states.append(var.state)
        return states

    @property
    def deriv(self):
        return self._get_deriv()

    def _get_deriv(self):
        derivs = []
        for var in self.state_var_list:
            derivs.append(var.deriv)
        return derivs


class DynSystem(BaseSystem):
    def __init__(self, initial_state, deriv_fun=None, output_fun=None):
        super().__init__(initial_state)

        if output_fun is None:
            def output_fun(t, x):
                return x

        self.initial_state = initial_state
        if isinstance(deriv_fun, SimObject) or isinstance(deriv_fun, BaseFunction):
            self.deriv_fun = deriv_fun.forward
        else:
            self.deriv_fun = deriv_fun

        self.output_fun = output_fun

    # override
    def reset(self):
        super().reset()
        self.state_var_list[0].apply_state(self.initial_state)
        return

    # override
    def set_state(self, state):
        self.state_var_list[0].apply_state(state)

    # to be implemented
    def derivative(self, time, state, *args, **kwargs):
        assert self.deriv_fun is not None, \
            "Attach a deriv_fun or implement the derivative method!"
        return self.deriv_fun(time, state, *args, **kwargs)

    # implement
    def forward(self, *args, **kwargs):
        deriv = self.derivative(
            self.sim_clock.time,
            self.state_var_list[0].state,
            *args, **kwargs
        )
        self.state_var_list[0].forward(deriv)

        if self.log_timer.is_event:
            self.logger.append(
                time=self.sim_clock.time,
                state=self.state_var_list[0].state)
            anonymous_values = {'u_' + str(i): value for i, value in enumerate(args)}
            self.logger.append(**anonymous_values)
            self.logger.append(**kwargs)

        return self.output

    # implement
    @property
    def output(self):
        return self.output_fun(
            self.sim_clock.time, self.state_var_list[0].state)

    # override
    def _get_state(self):
        return self.state_var_list[0].state

    # override
    def _get_deriv(self):
        return self.state_var_list[0].deriv


    @staticmethod
    def test():
        print("== Test for DynSystem ==")
        dt = 0.01
        sim_clock = SimClock()
        log_timer = Timer(dt)
        log_timer.attach_sim_clock(sim_clock)
        log_timer.turn_on()

        def deriv_fun(t, x):
            return -1/((1 + t)**2)*x

        model = DynSystem(1, deriv_fun)
        model.attach_sim_clock(sim_clock)
        model.attach_log_timer(log_timer)

        t = TicToc()
        t.tic()
        model.propagate(dt, 10)
        t.toc()
        model.plot()

        plt.show()


class TimeInvarDynSystem(DynSystem):
    def __init__(self, initial_state, deriv_fun=None, output_fun=None):
        if output_fun is None:
            def output_fun(x):
                return x
        super().__init__(initial_state, deriv_fun, output_fun)

    # to be implemented
    def derivative(self, state, *args, **kwargs):
        assert self.deriv_fun is not None, \
            "Attach a deriv_fun or implement the derivative method!"
        return self.deriv_fun(state, *args, **kwargs)

    # override
    def forward(self, *args, **kwargs):
        deriv = self.derivative(
            self.state_var_list[0].state,
            *args, **kwargs
        )
        self.state_var_list[0].forward(deriv)

        if self.log_timer.is_event:
            self.logger.append(
                time=self.sim_clock.time,
                state=self.state_var_list[0].state
            )
            anonymous_values = {'u_' + str(i): value for i, value in enumerate(args)}
            self.logger.append(**anonymous_values)
            self.logger.append(**kwargs)

        return self.output

    # override
    @property
    def output(self):
        return self.output_fun(
            self.state_var_list[0].state
        )

    @staticmethod
    def test():
        print("== Test for TimeInvarDynSystem ==")
        dt = 0.01
        sim_clock = SimClock()
        log_timer = Timer(dt)
        log_timer.attach_sim_clock(sim_clock)
        log_timer.turn_on()

        def deriv_fun(x, u):
            A = np.array([[0, 1], [-1, -1]])
            B = np.array([0, 1])
            return A.dot(x) + B.dot(u)

        model = TimeInvarDynSystem([0, 1], deriv_fun)
        model.attach_sim_clock(sim_clock)
        model.attach_log_timer(log_timer)

        t = TicToc()
        t.tic()
        model.propagate(dt, 10, u=1)
        t.toc()
        model.plot()

        plt.show()


class MultiStateDynSystem(BaseSystem):
    def __init__(self, initial_states, deriv_fun=None, output_fun=None):
        """ initial_states: tuple (state1, state2, ...) """
        if output_fun is None:
            def output_fun(*args):
                return args
        super(MultiStateDynSystem, self).__init__(*initial_states)
        self.name = "multi_state_dyn_system"
        self.initial_states = initial_states

        if isinstance(deriv_fun, SimObject) or isinstance(deriv_fun, BaseFunction):
            self.deriv_fun = deriv_fun.forward
        else:
            self.deriv_fun = deriv_fun

        self.output_fun = output_fun

    # override
    def reset(self):
        super(MultiStateDynSystem, self).reset()
        self.set_state(self.initial_states)

    # to be implemented
    def derivative(self, *args, **kwargs):
        """ implement this method if needed
        *args: [state1, state2, ..., input1, ...]
        return: [derivState1, derivState2, ...]
        """
        assert self.deriv_fun is not None,\
            "Attach a deriv_fun or implement the derivative method"
        return self.deriv_fun(*args, **kwargs)

    # implement
    def forward(self, *args, **kwargs):
        states = self._get_state()
        derivs = self.derivative(*states, *args, **kwargs)

        for i, state_var in enumerate(self.state_var_list):
            state_var.forward(derivs[i])

        if self.log_timer.is_event:
            self.logger.append(time=self.sim_clock.time)
            state_values = {'state_' + str(i): value for i, value in enumerate(states)}
            anonymous_values = {'u_' + str(i): value for i, value in enumerate(args)}
            self.logger.append(**state_values, **anonymous_values, **kwargs)

        return self.output

    # implement
    @property
    def output(self):
        states = self._get_state()
        return self.output_fun(*states)


class MultipleSystem(BaseSystem):
    def __init__(self):
        super(MultipleSystem, self).__init__()
        self.name = "multiple_system"
        self.sim_obj_list = []
        self.sim_obj_num = 0

    def attach_sim_objects(self, sim_obj_list):
        svi = self.state_var_num  # last state var index
        si = self.state_num  # last state index

        for sim_obj in sim_obj_list:
            if isinstance(sim_obj, SimObject):
                self.sim_obj_list.append(sim_obj)
                self.sim_obj_num += 1

                if isinstance(sim_obj, BaseSystem):
                    self.state_var_list.extend(sim_obj.state_var_list)
                    for index in sim_obj.state_index:
                        new_index = [si + _i for _i in index]
                        self.state_index.append(new_index)

                    svi += sim_obj.state_var_num
                    si += sim_obj.state_num
        self.state_var_num = svi
        self.state_num = si

    # override
    def attach_sim_clock(self, sim_clock):
        super(MultipleSystem, self).attach_sim_clock(sim_clock)
        for sim_obj in self.sim_obj_list:
            sim_obj.attach_sim_clock(sim_clock)

    # override
    def attach_log_timer(self, log_timer):
        super(MultipleSystem, self).attach_log_timer(log_timer)
        for sim_obj in self.sim_obj_list:
            sim_obj.attach_log_timer(log_timer)

    # override
    def reset(self):
        super(MultipleSystem, self).reset()
        for sim_obj in self.sim_obj_list:
            sim_obj.reset()

    # implement
    def check_stop_condition(self):
        to_stop_list = []
        for sim_obj in self.sim_obj_list:
            to_stop_list.append(sim_obj.check_stop_condition())

        to_stop = np.array(to_stop_list, dtype=bool).any()
        if to_stop:
            self.flag = []
            for sim_obj in self.sim_obj_list:
                self.flag.append(sim_obj.flag)

        return to_stop, self.flag


if __name__ == "__main__":
    DynSystem.test()
    TimeInvarDynSystem.test()


