import numpy as np
import matplotlib.pyplot as plt
import os
import h5py
from pytictoc import TicToc


class SimClock:
    def __init__(self, time=0, time_res=1e-6):
        self.time = time  # Current time
        self.time_res = time_res  # Time resolution

    def reset(self):
        self.time = 0

    def apply_time(self, time):
        self.time = time

    def apply_time_res(self, time_res):
        self.time_res = time_res

    def elapse(self, dt):
        self.time += dt


class Timer:
    def __init__(self, event_time_interval=0.0):
        self.sim_clock = None
        self.is_operating = False
        self.event_time_interval = event_time_interval
        self.last_event_time = -np.inf
        self.is_event = False

    def attach_sim_clock(self, sim_clock):
        self.sim_clock = sim_clock

    def reset(self):
        self.is_operating = False
        self.last_event_time = -np.inf
        self.is_event = False

    def turn_on(self, event_time_interval=None):
        if event_time_interval is not None:
            self.event_time_interval = event_time_interval
        self.is_operating = True
        self.last_event_time = self.sim_clock.time
        self.is_event = True

    def turn_off(self):
        self.is_operating = False
        self.last_event_time = -np.inf
        self.is_event = False

    def forward(self):
        if abs(self.sim_clock.time - self.last_event_time) <= self.sim_clock.time_res:
            return

        if self.is_operating:
            elapsed_time = self.sim_clock.time - self.last_event_time
            if elapsed_time >= (self.event_time_interval - self.sim_clock.time_res):
                self.is_event = True
                self.last_event_time = self.sim_clock.time
            else:
                self.is_event = False

    @staticmethod
    def test():
        print("== Test for Timer ==")

        dt = 0.01
        sim_clock = SimClock()

        event_time_interval = 0.1
        timer = Timer(event_time_interval)
        timer.attach_sim_clock(sim_clock)
        timer.turn_on()

        print("Initial time: 0.0[s]")
        print("Sampling time interval: {:.2f}[s]".format(dt))
        print("Event time interval: {:.2f}[s]\n".format(event_time_interval))
        for i in range(100):
            timer.forward()
            if timer.is_event:
                print("[Timer] An event occurred at time = {:.1f}[s]".format(sim_clock.time))
            sim_clock.elapse(dt)


class Logger:
    def __init__(self):
        self.data = dict()
        self.is_operating = True
        self.name = "Logger"

    def turn_on(self):
        self.is_operating = True

    def turn_off(self):
        self.is_operating = False

    def clear(self):
        self.data = dict()

    def is_empty(self):
        return self.data is None

    def len(self):
        if self.data is None:
            return 0
        else:
            first_key = next(iter(self.data.keys()))
            return len(self.data[first_key])

    def append(self, **kwargs):
        if self.is_operating:
            for key in kwargs.keys():
                try:
                    self.data[key].append(kwargs[key])
                except KeyError:
                    self.data[key] = []
                    self.data[key].append(kwargs[key])

    def get(self, *keys):
        if self.is_empty():
            return None

        if len(keys) == 0:
            keys = self.data.keys()

        if len(keys) == 1:
            return np.array(self.data[keys[0]])
        else:
            return {key: np.array(self.data[key]) for key in keys}

    def keys(self):
        return self.data.keys()

    def save(self, h5file=None, data_group='/'):
        if h5file is None:
            if not os.path.isdir('./data'):
                os.mkdir('./data')
            h5file = h5py.File('./data/log.hdf5', 'w')

        logged_data = self.get()
        with h5file:
            for key, value in logged_data.items():
                h5file[data_group + key] = value

    def load(self, h5file=None, data_group='/'):
        self.clear()
        if h5file is None:
            h5file = h5py.File('./data/log.hdf5', 'r')

        with h5file:
            for key, value in h5file[data_group].items():
                self.data[key] = list(value)

    @staticmethod
    def test():
        print("== Test for Logger ==")

        dt = 0.01
        sim_clock = SimClock()
        logger = Logger()

        A = np.array(
            [[1, dt], [0, 1]]
        )
        B = np.array([0, dt])
        x = np.array([0, 0])
        u = 1

        t = TicToc()
        t.tic()
        for i in range(100):
            logger.append(time=sim_clock.time, state=x, control=u)
            x = A.dot(x) + B.dot(u)
            sim_clock.elapse(dt)
        t.toc()

        logged_data = logger.get()
        plt.figure()
        plt.plot(logged_data['time'], logged_data['state'], label={"Pos. [m]", "Vel. [m/s]"})
        plt.xlabel("Time [s]")
        plt.ylabel("State")
        plt.grid()
        plt.legend()
        plt.show()


if __name__ == "__main__":
    Timer.test()
    Logger.test()

