import numpy as np
import matplotlib.pyplot as plt
from pysimenv.core.simulator import Simulator
from pysimenv.core.systems import TimeInvarDynSystem
from pysimenv.core.functions import BaseFunction
from pysimenv.common.models import FeedbackControl

"""
Reference: B. L. Stevens, F. L. Lewis, E. N. Johnson, "Aircraft Control and Simulation", 3rd edition
Example 5.3-1: LQR Design for F-16 Lateral Regulator (pp.407)
State:
    x = [beta, phi, p, r, delta_a, delta_r, r_w]
    beta, phi: [rad], p, r: [rad/s], delta_a, delta_r: [deg], r_w: [deg]
Input:
    u = [u_a, u_r]
    u_a: aileron servo input [deg]
    u_r: rudder servo input [deg]
Output:
    y = [r_w, p, beta, phi]
    r_w, beta, phi: [deg], p: [deg/s]
"""
A = np.array([
    [-0.3220, 0.0640, 0.0364, -0.9917, 0.0003, 0.0008, 0],
    [0, 0, 1, 0.0037, 0, 0, 0],
    [-30.6492, 0, -3.6784, 0.6646, -0.7333, 0.1315, 0],
    [8.5396, 0, -0.0254, -0.4764, -0.0319, -0.0620, 0],
    [0, 0, 0, 0, -20.2, 0, 0],
    [0, 0, 0, 0, 0, -20.2, 0],
    [0, 0, 0, 57.2958, 0, 0, -1]
])
B = np.array([
    [0, 0],
    [0, 0],
    [0, 0],
    [0, 0],
    [20.2, 0],
    [0, 20.2],
    [0, 0]
])
C = np.array([
    [0, 0, 0, 57.2958, 0, 0, -1],
    [0, 0, 57.5928, 0, 0, 0, 0],
    [57.5928, 0, 0, 0, 0, 0, 0],
    [0, 57.5928, 0, 0, 0, 0, 0]
])
K = np.array([
    [-0.56, -0.44, 0.11, -0.35],
    [-1.19, -0.21, -0.44, 0.26]])

linear_system = TimeInvarDynSystem(
    initial_state=[1, 0, 0, 0, 0, 0, 0],
    deriv_fun=lambda x, u: A.dot(x) + B.dot(u),
    output_fun=lambda x: C.dot(x))
control = BaseFunction(lambda y, r: -K.dot(y))
model = FeedbackControl(linear_system, control)  # closed-loop system

simulator = Simulator(model)
simulator.propagate(dt=0.01, time=10)
linear_system.plot(
    var_ind_dict={'state': [0, 3, 1, 2]},
    var_names_dict={'state': ['beta', 'r', 'phi', 'p'], 'u_0': ['u_a', 'u_r']}
)

plt.show()
