import numpy as np
import matplotlib.pyplot as plt
from pysimenv.core.systems import TimeInvarDynSystem, MultipleSystem
from pysimenv.common.common_utils import wrap_to_pi
from pysimenv.missile.missile_utils import RelKin2dim, CloseDistCond, miss_distance
from pysimenv.missile.guidance import PurePNG2dim


class PlanarManVehicle2dof(TimeInvarDynSystem):
    """
    A 2-dof model for a maneuvering vehicle in the plane
    state x: [p_x, p_y, V, gamma]
    control u: [a_x, a_y]
    where [p_x, p_y] is the position expressed in the inertial frame and
    [a_x, a_y] is the acceleration expressed in the velocity frame
    """
    def __init__(self, initial_state):
        super(PlanarManVehicle2dof, self).__init__(initial_state)
        self.name = "man_vehicle"

        def angle_correction_fun(state):
            state[3] = wrap_to_pi(state[3])  # gamma should be in [-pi, pi] rad
            return state
        self.state_var_list[0].attach_correction_fun(angle_correction_fun)

    # implement
    def derivative(self, x, u):
        """
        :param x: [p_x, p_y, V, gamma]
        :param u: [a_x, a_y]
        :return: x_dot
        """
        V = x[2]
        gamma = x[3]

        a_x = u[0]
        a_y = u[1]

        p_x_dot = V*np.cos(gamma)
        p_y_dot = V*np.sin(gamma)
        V_dot = a_x

        if abs(V) < 1e-4:
            gamma_dot = 0
        else:
            gamma_dot = a_y/V

        return np.array([p_x_dot, p_y_dot, V_dot, gamma_dot])

    @property
    def p(self):
        """
        :return: [p_x, p_y]
        """
        return self.state[0:2]

    @property
    def v(self):
        """
        :return: [v_x, v_y]
        """
        x = self.state
        V = x[2]
        gamma = x[3]

        return np.array([[V*np.cos(gamma), V*np.sin(gamma)]])

    @property
    def V(self):
        """
        :return: V
        """
        return self.state[2]

    @property
    def gamma(self):
        """
        :return: gamma
        """
        return self.state[3]

    # override
    def plot(self):
        var_keys = {'state', 'u_0'}
        var_ind_dict = {'state': [0, 1, 2, 3], 'u_0': [0, 1]}
        var_names_dict = {
            'state': ['p_x (m)', 'p_y (m)', 'V (m/s)', 'gamma (rad)'],
            'u_0': ['a_x (m/s**2)', 'a_y (m/s**2)']
        }
        super(PlanarManVehicle2dof, self).plot(var_keys, var_ind_dict, var_names_dict)

    def plot_path(self, fig_ax=None):
        if fig_ax is None:
            fig, ax = plt.subplots()
            ax.set_xlabel("p_x (m)")
            ax.set_ylabel("p_y (m)")
            ax.grid()
        else:
            fig = fig_ax['fig']
            ax = fig_ax['ax']

        state_list = self.history('state')
        pos_list = state_list[:, 0:2]
        ax.plot(pos_list[:, 0], pos_list[:, 1], label=self.name + " Traj.")
        ax.set_aspect('equal')
        ax.legend()

        plt.draw()
        plt.pause(0.01)
        return {'fig': fig, 'ax': ax}


class PlanarNonManVehicle2dof(PlanarManVehicle2dof):
    def __init__(self, initial_state):
        super(PlanarNonManVehicle2dof, self).__init__(initial_state)
        self.name = "non_man_vehicle"

    # override
    def derivative(self, x):
        return super(PlanarNonManVehicle2dof, self).derivative(x, np.zeros(2))

    # override
    def plot(self):
        var_keys = {'state'}
        var_ind_dict = {'state': [0, 1, 2, 3]}
        var_names_dict = {
            'state': ['p_x (m)', 'p_y (m)', 'V (m/s)', 'gamma (rad)'],
        }
        super(PlanarManVehicle2dof, self).plot(var_keys, var_ind_dict, var_names_dict)


class PlanarMissile2dof(PlanarManVehicle2dof):
    def __init__(self, initial_state):
        super(PlanarMissile2dof, self).__init__(initial_state)
        self.name = "missile"
        self.fov_limit = np.inf
        self.acc_limit = np.array([
            [-np.inf, np.inf],
            [-np.inf, np.inf]
        ])
        self.ground_elevation = 0

    # override
    def forward(self, a_M):
        a_M = np.clip(a_M, self.acc_limit[:, 0], self.acc_limit[:, 1])
        super(PlanarMissile2dof, self).forward(a_M)

    # implement
    def check_stop_condition(self):
        to_stop = False
        if self.is_stalled():
            to_stop = True
            self.flag = 1
        if self.is_collided():
            to_stop = True
            self.flag = 2
        return to_stop, self.flag

    def is_stalled(self):
        return self.V < 10  # when the speed is less than 10m/s

    def is_collided(self):
        return self.p[1] < self.ground_elevation - 0.5

    def report(self):
        np.set_printoptions(precision=2, suppress=True)
        print("[{:s}] Final state [m, m, m/s, deg]:".format(self.name))

        final_state = self.state
        final_state[3] = np.rad2deg(final_state[3])
        print(final_state)
        if self.flag == 0:
            print("[{:s}] Status: normal".format(self.name))
        elif self.flag == 1:
            print("[{:s}] Status: stalled".format(self.name))
        elif self.flag == 2:
            print("[{:s}] Status: collided".format(self.name))


class Engagement2dim(MultipleSystem):
    def __init__(self, missile, target):
        super(Engagement2dim, self).__init__()
        self.missile = missile
        self.target = target
        self.relKin = RelKin2dim()
        self.close_dist_cond = CloseDistCond(r_threshold=10.0)

        self.attach_sim_objects([self.missile, self.target])

        x_M = self.missile.state
        x_T = self.target.state
        self.relKin.forward(x_M, x_T)

    # override
    def reset(self):
        super(Engagement2dim, self).reset()
        self.close_dist_cond.reset()

    # implement
    def forward(self):
        x_M = self.missile.state
        x_T = self.target.state
        self.relKin.forward(x_M, x_T)
        self.close_dist_cond.forward(self.relKin.r)
        self.target.forward()

    # implement
    def check_stop_condition(self):
        to_stop = False

        to_stop_, _ = self.missile.check_stop_condition()
        if to_stop_:  # stop due to the missile
            to_stop = True
            self.flag = 1

        if self.close_dist_cond.check():  # probable interception
            to_stop = True
            self.flag = 2
        return to_stop, self.flag

    def miss_distance(self):
        x_M = self.missile.history('state')
        x_T = self.missile.history('state')

        p_M = x_M[:, 0:2]
        p_T = x_T[:, 0:2]
        d_miss = miss_distance(p_M, p_T)
        return d_miss

    def report(self):
        self.missile.report()
        print("[Engagement] Miss distance: {:.4f} (m)".format(self.miss_distance()))

    def plot(self):
        self.missile.plot()

        figs = dict()

        fig_ax = self.missile.plot_path()
        fig1, ax1 = fig_ax['fig'], fig_ax['ax']
        self.target.plot_path({'fig': fig1, 'ax': ax1})
        fig1.suptitle("2-dim flight path")
        figs['path'] = {'fig': fig1, 'ax': ax1}

        # plots for rel. kin.
        time_list = self.missile.history('time')
        x_M_list = self.missile.history('state')
        x_T_list = self.target.history('state')

        relKin = RelKin2dim()
        r_list = []
        sigma_list = []
        lam_list = []
        omega_list = []

        zem_list = []

        for i in range(x_M_list.shape[0]):
            x_M = x_M_list[i, :]
            x_T = x_T_list[i, :]

            relKin.forward(x_M, x_T)

            r_list.append(relKin.r)
            sigma_list.append(relKin.sigma)
            lam_list.append(relKin.lam)
            omega_list.append(relKin.omega)

            zem_list.append(relKin.zem)

        r_list = np.array(r_list)
        sigma_list = np.array(sigma_list)
        lam_list = np.array(lam_list)
        omega_list = np.array(omega_list)

        zem_list = np.array(zem_list)

        fig2, ax2 = plt.subplots(2, 2)
        fig2.tight_layout()
        ax2_1 = ax2[0, 0]
        ax2_1.set_title("Rel. dist")
        ax2_1.plot(time_list[:-1], r_list[:-1], label="Rel. dist")
        ax2_1.set_xlabel("Time (s)")
        ax2_1.set_ylabel("Rel. dist (m)")
        ax2_1.grid()
        ax2_1.legend()

        ax2_2 = ax2[0, 1]
        ax2_2.set_title("Look angle")
        ax2_2.plot(time_list[:-1], np.rad2deg(sigma_list[:-1]), label="look angle")
        ax2_2.set_xlabel("Time (s)")
        ax2_2.set_ylabel("Look angle (deg)")
        ax2_2.grid()
        ax2_2.legend()

        ax2_3 = ax2[1, 0]
        ax2_3.set_title("LOS angle")
        ax2_3.plot(time_list[:-1], np.rad2deg(lam_list[:-1]), label="LOS angle")
        ax2_3.set_xlabel("Time (s)")
        ax2_3.set_ylabel("LOS angle (deg)")
        ax2_3.grid()
        ax2_3.legend()

        ax2_4 = ax2[1, 1]
        ax2_4.set_title("LOS rate")
        ax2_4.plot(time_list[:-1], np.rad2deg(omega_list[:-1]), label="LOS rate")
        ax2_4.set_xlabel("Time (s)")
        ax2_4.set_ylabel("LOS rate (deg/s)")
        ax2_4.grid()
        ax2_4.legend()

        figs['Rel. Kin.'] = {'fig': fig2, 'ax': ax2}

        fig3, ax3 = plt.subplots()
        ax3.set_title("ZEM")
        ax3.plot(time_list[:-1], zem_list[:-1], label="ZEM")
        ax3.set_xlabel("Time (s)")
        ax3.set_ylabel("ZEM (m)")
        ax3.grid()
        ax3.legend()

        figs['Rel. Kin. add.'] = {'fig': fig3, 'ax': ax3}

        return figs


class PurePNG2dimEngagement(Engagement2dim):
    def __init__(self, missile, target):
        super(PurePNG2dimEngagement, self).__init__(missile, target)

        self.pure_png = PurePNG2dim(N=3.0)

    # implement
    def forward(self):
        super(PurePNG2dimEngagement, self).forward()
        V_M = self.missile.V
        omega = self.relKin.omega

        a_M = self.pure_png.forward(V_M, omega)
        self.missile.forward(np.array([0, a_M]))
